/*
 * Copyright 2009 - 2010 by Ping Cheng, Wacom. <pingc@wacom.com>
 * Copyright 2011 by Alexey Osipov. <simba@lerlan.ru>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef __XF86_WCMTOUCHFILTER_H
#define __XF86_WCMTOUCHFILTER_H

#include "xf86Wacom.h"

/****************************************************************************/

void wcmGestureFilter(WacomDevicePtr priv, int touch_id);
Bool wcmTouchNeedSendEvents(WacomCommonPtr common);

/****************************************************************************/
#endif /* __XF86_WCMTOUCHFILTER_H */
